/*******************************************************************************/
/* Elevator model                                                              */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* modelled after                                                              */
/*     Meinke, K., Sindhu, M.A.: Incremental learning-based testing for 
       reactive systems. In: TAP. Volume 6706 of LNCS. (2011) 134–151          */
/* if you reuse this code, please cite also                                    */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <assert.h>

typedef struct state {
  int stop;
  int cl;
  int w1;
  int w2;
  int w3;
  int at;
  int up;
  int sw2; //stop at 2
} t_state;

typedef struct io {
  int tick; 
  int c1; 
  int c2; 
  int c3; 
} t_io;

void init(t_state *s) {
  s->stop = 1;
  s->cl = 1;
  s->w1 = 0;
  s->w2 = 0;
  s->w3 = 0;
  s->at = 1;
  s->up = 1;
  s->sw2 = 0;
}

void compute(t_io* i, t_state *s) {
  if(s->stop) {
    if(!s->w1 && s->at==1) {
      if(!s->w2 && !s->w3) {
        if(s->cl && i->c1) { s->cl = 0; return;}
        if(i->c2) { s->w2 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 0; return; }
        if(!s->cl && i->tick) { s->cl = 1; return; }
      }
      if(!s->w2 && s->w3 && !s->cl) {
        if(i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 0; return; }
        if(i->c2) { s->w2 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->w2 && !s->w3 && !s->cl) {
        if(i->c2 || i->tick) { s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 0; return; }
      }
      if(s->w2 && s->w3 && !s->cl) {
        if(i->c2 || i->c3 || i->tick) { s->cl = 1; s->stop = 0; s->up = 1; s->at = 0; s->sw2 = 1; return; }
      }
    }
    else if(!s->w2 && s->at==2) {
      if(!s->w1 && !s->w3) {
        if(s->cl && i->c2) { s->cl = 0; return; }
        if(i->c1) { s->w1 = 1; s->cl = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->cl = 1; s->stop = 0; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(!s->cl && i->tick) { s->cl = 1; return; }
      }
      if(s->up && !s->w1 && s->w3 && !s->cl) {
        if(i->c1) { s->w1 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(!s->up && !s->w1 && s->w3 && !s->cl) {
        if(i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->up = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c1) { s->w1 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->up && s->w1 && !s->w3 && !s->cl) {
        if(i->c1 || i->tick) { s->stop = 0; s->cl = 1; s->up = 0; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(!s->up && s->w1 && !s->w3 && !s->cl) {
        if(i->c1 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
        if(i->c3) { s->w3 = 1; s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->up && s->w1 && s->w3) {
        if(i->c1 || i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(!s->up && s->w1 && s->w3) {
        if(i->c1 || i->c3 || i->tick) { s->stop = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
    }
    else if(!s->w3 && s->at==3) {
      if(!s->w1 && !s->w2) {
        if(s->cl && i->c3) { s->cl = 0; return; }
        if(i->c1) { s->w1 = 1; s->cl = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 0; return; }
        if(i->c2) { s->w2 = 1; s->cl = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 1; return; }
	if(!s->cl && i->tick) { s->cl = 1; return; }
      }
      if(!s->w1 && s->w2 && !s->cl) {
        if(i->c1) { s->w1 = 1; s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 0; return; }
        if(i->c2 || i->tick) { s->stop = 0; s->up = 0; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->w1 && !s->w2 && !s->cl) {
        if(i->c1 || i->tick) { s->stop = 0; s->up = 0; s->cl = 1; s->at = 0; s->sw2 = 0; return; }
        if(i->c2) { s->w2 = 1; s->stop = 0; s->up = 0; s->cl = 1; s->at = 0; s->sw2 = 1; return; }
      }
      if(s->w1 && s->w2) {
        if(i->c1 || i->c2 || i->tick) { s->cl = 1; s->stop = 0; s->up = 0; s->sw2 = 1; return; }
      }
    }
  }
  else if(s->cl && s->at!=1 && s->at!=2 && s->at!=3) { //go
    if(s->up) { //up
      if(!s->sw2) {
        if(!s->w2 && s->w3) {
          if(!s->w1) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
            if(i->c3 || i->tick) { s->sw2 = 1; return; }
  	  }
          if(s->w1) {
            if(i->c1 || i->c3 || i->tick) { s->sw2 = 1; return; }
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
  	  }
        }
      }
      else {
	if(s->w2) {
	  if(s->w1 && !s->w3) {
            if(i->c3) { s->w3 = 1; return; }
            if(i->c2 || i->tick) { s->w1 = 1; s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(s->w1 && s->w3) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && !s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->tick) {  s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	}
	if(s->w3) {
	  if(s->w1 && !s->w2) {
            if(i->c2) { s->w2 = 1; return; }
            if(i->c1 || i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; s->cl = 0; return; }
	  }
	  if(s->w1 && s->w2) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; return; }
	  }
	  if(!s->w1 && !s->w2) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2) { s->w2 = 1; return; }
            if(i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && s->w2) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->c3 || i->tick) { s->w3 = 0; s->at = 3; s->stop = 1; s->cl = 0; return; }
	  }      
	}
      }
    }
    else { //down
      if(!s->sw2) {
	if(s->w1 && !s->w2) {
	  if(!s->w3) {
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->tick) { s->sw2 = 1; return; }
	  }
	  if(s->w3) {
            if(i->c2) { s->w2 = 1; s->sw2 = 1; return; }
            if(i->c1 || i->c3 || i->tick) { s->sw2 = 1; return; }
	  }
	}
      }
      else {
	if(s->w2) {
	  if(s->w1 && !s->w3) {
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->c2 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(s->w1 && s->w3) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && !s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c3) { s->w3 = 1; return; }
            if(i->c2 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w1 && s->w3) {
            if(i->c1) { s->w1 = 1; return; }
            if(i->c2 || i->c3 || i->tick) { s->w2 = 0; s->at = 2; s->stop = 1; s->cl = 0; return; }
	  }
	}
	if(s->w1) {
	  if(s->w2 && !s->w3) {
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->c2 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; } 
	  }
	  if(s->w2 && s->w3) {
            if(i->c1 || i->c2 || i->c3 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w2 && !s->w3) {
            if(i->c2) { s->w2 = 1; return; }
            if(i->c3) { s->w3 = 1; return; }
            if(i->c1 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; }
	  }
	  if(!s->w2 && s->w3) {
            if(i->c2) { s->w2 = 1; return; }
            if(i->c1 || i->c3 || i->tick) { s->w1 = 0; s->at = 1; s->stop = 1; s->cl = 0; return; }
	  }      
	}
      }
    }
  }
}

#ifndef NO_CBMC

void state_assume(t_state* s) {
  __CPROVER_assume((0<=s->stop) && (s->stop<=1) &&
                   (0<=s->cl) && (s->cl<=1) &&
                   (0<=s->w1) && (s->w1<=1) &&
                   (0<=s->w2) && (s->w2<=1) &&
                   (0<=s->w3) && (s->w3<=1) &&
                   (1<=s->at) && (s->at<=3) &&
  (s->stop && 
    (!s->w1 && s->at==1 &&
      (!s->w2 && !s->w3 || !s->w2 && s->w3 && !s->cl ||
       s->w2 && !s->w3 && !s->cl || s->w2 && s->w3 && !s->cl))
    || 
    (!s->w2 && s->at==2 &&
      (!s->w1 && !s->w3 || !s->w1 && s->w3 && !s->cl ||
       s->w1 && !s->w3 && !s->cl || s->w1 && s->w3 && !s->cl))
    || 
    (!s->w3 && s->at==3 &&
      (!s->w1 && !s->w2 || !s->w1 && s->w2 && !s->cl ||
       s->w1 && !s->w2 && !s->cl || s->w1 && s->w2 && !s->cl)) 
   )
  || !s->stop && s->cl && s->at==0 &&
      (s->up && s->sw2 || s->up && !s->sw2 && !s->w2 && s->w3 ||
       !s->up && s->sw2 || !s->up && !s->sw2 && !s->w2 && s->w1)
);
}

void input_assume(t_io* i) {
  __CPROVER_assume((0<=i->tick) && (i->tick<=1) &&
                   (0<=i->c1) && (i->c1<=1) &&
                   (0<=i->c2) && (i->c2<=1) &&
                   (0<=i->c3) && (i->c3<=1) &&
                   (
                     (i->c1 || i->c2 || i->c3) && !i->tick ||
                     !i->c1 && !i->c2 && !i->c3 && i->tick
		    )
                  );
}

void final(t_io* i, t_state* s) {
  __CPROVER_assume(s->stop && (s->at==1) && s->cl && !s->w1 && !s->w2 && !s->w3);
}

//stopped, no request => doors closed
void prop1(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->cl && !s->w1 && !s->w2 && !s->w3 && !i->c1 && !i->c2 && !i->c3);
  //step
  compute(i,s);
  //postcondition
  assert(s->cl);
}

//stopped, doors open, any request => doors close
void prop2(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->cl && !s->w1 && !s->w2 && !s->w3 && 
    (s->at==1 && !i->c1 && (i->c2 || i->c3) ||
     s->at==2 && !i->c2 && (i->c1 || i->c3) ||
     s->at==3 && !i->c3 && (i->c1 || i->c2))
    );
  //step
  compute(i,s);
  //postcondition
  assert(!s->stop && s->up);
}

//stopped, doors closed, request at respective floor => doors open
void prop3(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->cl && !s->w1 && !s->w2 && !s->w3 && 
		   (s->at==1 && i->c1 || s->at==2 && i->c2 || s->at==3 && i->c3));
  //step
  compute(i,s);
  //postcondition
  assert(!s->cl);
}

//arriving at requested floor => stop and doors open
void prop4(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && i->tick && 
     (s->w1 && !s->w2 && !s->w3 && s->sw2 && !s->up  ||
      !s->w1 && s->w2 && !s->w3 && s->sw2  ||
      !s->w1 && !s->w2 && s->w3 && s->sw2 && s->up) 
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl);
}

//not stopped, request => request stored
void prop5(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && !s->w1 && i->c1 
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w1);
}

//not stopped, request => request stored
void prop6(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && !s->w2 && i->c2
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w2);
}

//not stopped, request => request stored
void prop7(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && !s->w3 && i->c3
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w3);
}

//doors open, request at respective floor => doors open
void prop8(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->cl && s->at==1 && i->c1);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && !s->w1);
}

//doors open, request at respective floor => doors open
void prop9(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->cl && s->at==2 && i->c2);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && !s->w2);
}

//doors open, request at respective floor => doors open
void prop10(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->cl && s->at==3 && i->c3);
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && !s->w3);
}

//stopped, doors open, any request => doors close
void prop11(t_io *i, t_state* s) {
  //precondition
    __CPROVER_assume(s->stop && s->at==1 && !s->cl && !s->w1 && !s->w2 && !s->w3 && !i->c1 && (i->c2 || !i->c3));
  //step
  compute(i,s);
  //postcondition
  assert(s->cl);
}

//stopped, doors open, any request => doors close
void prop12(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->at==2 && !s->cl && !s->w1 && !s->w2 && !s->w3 && (i->c1 || i->c3) && !i->c2);
  //step
  compute(i,s);
  //postcondition
  assert(s->cl);
}

//stopped, doors open, any request => doors close
void prop13(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->at==3 && !s->cl && !s->w1 && !s->w2 && !s->w3 && (i->c1 || i->c2) && !i->c3);
  //step
  compute(i,s);
  //postcondition
  assert(s->cl);
}

//stopped, doors closed, request at respective floor => doors open
void prop14(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->at==1 && s->cl && !s->w1 && !s->w2 && !s->w3 && i->c1);
  //step
  compute(i,s);
  //postcondition
  assert(!s->cl);
}

//stopped, doors closed, request at respective floor => doors open
void prop15(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->at==2 && s->cl && !s->w1 && !s->w2 && !s->w3 && i->c2);
  //step
  compute(i,s);
  //postcondition
  assert(!s->cl);
}

//stopped, doors closed, request at respective floor => doors open
void prop16(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->stop && s->at==3 && s->cl && !s->w1 && !s->w2 && !s->w3 && i->c3);
  //step
  compute(i,s);
  //postcondition
  assert(!s->cl);
}

//arriving at requested floor => stop and doors open
void prop17(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && !s->up &&
     s->w1 && !s->w2 && !s->w3 && s->sw2 && !i->c2 && !i->c3
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && s->at==1);
}

//arriving at requested floor => stop and doors open
void prop18(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && !s->up &&
     !s->w1 && s->w2 && !s->w3 && s->sw2 && !i->c1 && !i->c3
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && s->at==2);
}

//arriving at requested floor => stop and doors open
void prop19(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && s->up &&
     !s->w1 && s->w2 && !s->w3 && s->sw2 && !i->c1 && !i->c3
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && s->at==2);
}

//arriving at requested floor => stop and doors open
void prop20(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->cl && s->up &&
     !s->w1 && !s->w2 && s->w3 && s->sw2 && !i->c1 && !i->c2
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->stop && !s->cl && s->at==3);
}

//up, request => request stored
void prop21(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->up && s->cl && !s->w1 && i->c1
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w1);
}

//up, request => request stored
void prop22(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->up && s->cl && !s->w2 && i->c2
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w2);
}

//up, request => request stored
void prop23(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && s->up && s->cl && !s->w3 && i->c3
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w3);
}

//down, request => request stored
void prop24(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && !s->up && s->cl && !s->w1 && i->c1
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w1);
}

//down, request => request stored
void prop25(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && !s->up && s->cl && !s->w2 && i->c2
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w2);
}

//down, request => request stored
void prop26(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(!s->stop && !s->up && s->cl && !s->w3 && i->c3
    );
  //step
  compute(i,s);
  //postcondition
  assert(s->w3);
}

#endif


